package ru.dbotthepony.kstarbound.defs

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.*
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap

private fun flattenJsonPrimitive(input: JsonPrimitive): Any {
	if (input.isNumber) {
		return input.asNumber
	} else if (input.isString) {
		return input.asString.intern()
	} else {
		return input.asBoolean
	}
}

private fun flattenJsonArray(input: JsonArray): ArrayList<Any> {
	val flattened = ArrayList<Any>(input.size())

	for (v in input) {
		when (v) {
			is JsonObject -> flattened.add(flattenJsonObject(v))
			is JsonArray -> flattened.add(flattenJsonArray(v))
			is JsonPrimitive -> flattened.add(flattenJsonPrimitive(v))
			// is JsonNull -> baked.add(null)
		}
	}

	return flattened
}

private fun flattenJsonObject(input: JsonObject): Object2ObjectArrayMap<String, Any> {
	val flattened = Object2ObjectArrayMap<String, Any>()

	for ((k, v) in input.entrySet()) {
		when (v) {
			is JsonObject -> flattened[k] = flattenJsonObject(v)
			is JsonArray -> flattened[k] = flattenJsonArray(v)
			is JsonPrimitive -> flattened[k] = flattenJsonPrimitive(v)
		}
	}

	return flattened
}

fun flattenJsonElement(input: JsonElement): Any? {
	return when (input) {
		is JsonObject -> flattenJsonObject(input)
		is JsonArray -> flattenJsonArray(input)
		is JsonPrimitive -> flattenJsonPrimitive(input)
		is JsonNull -> null
		else -> throw IllegalArgumentException("Unknown argument $input")
	}
}

/**
 * Возвращает глубокую неизменяемую копию [input] примитивов/List'ов/Map'ов
 */
fun enrollList(input: List<Any>): ImmutableList<Any> {
	val builder = ImmutableList.builder<Any>()

	for (v in input) {
		when (v) {
			is Map<*, *> -> builder.add(enrollMap(v as Map<String, Any>))
			is List<*> -> builder.add(enrollList(v as List<Any>))
			else -> builder.add((v as? String)?.intern() ?: v)
		}
	}

	return builder.build()
}

/**
 * Возвращает глубокую неизменяемую копию [input] примитивов/List'ов/Map'ов
 */
fun enrollMap(input: Map<String, Any>): ImmutableMap<String, Any> {
	val builder = ImmutableMap.builder<String, Any>()

	for ((k, v) in input) {
		when (v) {
			is Map<*, *> -> builder.put(k.intern(), enrollMap(v as Map<String, Any>))
			is List<*> -> builder.put(k.intern(), enrollList(v as List<Any>))
			else -> builder.put(k.intern(), (v as? String)?.intern() ?: v)
		}
	}

	return builder.build()
}

/**
 * Возвращает глубокую изменяемую копию [input] примитивов/List'ов/Map'ов
 */
fun flattenList(input: List<Any>): ArrayList<Any> {
	val list = ArrayList<Any>(input.size)

	for (v in input) {
		when (v) {
			is Map<*, *> -> list.add(flattenMap(v as Map<String, Any>))
			is List<*> -> list.add(flattenList(v as List<Any>))
			else -> list.add(v)
		}
	}

	return list
}

fun flattenMap(input: Map<String, Any>): Object2ObjectArrayMap<String, Any> {
	val map = Object2ObjectArrayMap<String, Any>()

	for ((k, v) in input) {
		when (v) {
			is Map<*, *> -> map[k] = flattenMap(v as Map<String, Any>)
			is List<*> -> map[k] = flattenList(v as List<Any>)
			else -> map[k] = v
		}
	}

	return map
}

/**
 * Базовый класс описания прототипа игрового объекта
 *
 * Должен иметь все (или больше) поля объекта, который он будет создавать
 *
 * Поля должны иметь базовые ограничения (т.е. ограничения, которые применимы для всех конфигураций прототипа).
 * Если границы поля зависят от других полей, то проверка такого поля должна осуществляться уже при самой
 * сборке прототипа.
 */
abstract class ConfigurableDefinition<Configurable : ConfigurableDefinition<Configurable, Configured>, Configured : ConfiguredDefinition<Configured, Configurable>> {
	val json = Object2ObjectArrayMap<String, Any>()
	fun enroll() = enrollMap(json)
	abstract fun configure(directory: String = ""): Configured
}

/**
 * Базовый класс описанного прототипа игрового объекта
 *
 * Должен иметь все поля объекта, которые будут использоваться движком напрямую
 *
 * Создается соответствующим [ConfigurableDefinition], который проверил уже все поля
 * на их правильность.
 */
abstract class ConfiguredDefinition<Configured : ConfiguredDefinition<Configured, Configurator>, Configurator : ConfigurableDefinition<Configurator, Configured>>(
	val json: ImmutableMap<String, Any>
) {
	open fun getParameter(key: String): Any? = json[key]
	fun flatten() = flattenMap(json)
	abstract fun reconfigure(): Configurator
}
